(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/AbsynDot.sml,v
   Comment: Dumps hierarchy structure to dot file (for drawing)
   Release: scope-0-16_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   AbsynDot.sml,v 1.3 2003/07/08 15:45:01 wasowski Exp
*)

structure AbsynDot : CODGEN = struct

open Absyn CodGen Cppish

infix 0 << <<< <<> <- (* some fun with C++ style IO *)
infix 3 --

val --> = Dot.-->
infix 0 --> (* edge constructor for Dot *)

structure S = Symbol

type options = string option
type aor = unit
type atr = unit

fun simplename (s :S.symbol) :string=
	List.last (String.tokens (fn c => c = #".") (S.name s))

fun genname (s :S.symbol) :string =
	"node" ^ Int.toString (S.toInt s)

fun fo (ORState s, k :Dot.stmt list) :Dot.stmt list =
	let val this = genname (#name s)
	    val par  = genname (#parent s)
	in Dot.Node {id=this,attrs=
		[Dot.Label (simplename (#name s)), Dot.Style Dot.Dashed]} 
		:: (par --> this) [] :: k
	end

fun fa (ANDState s, k :Dot.stmt list) :Dot.stmt list =
	let val this = genname (#name s)
	    val n = Dot.Node {id=this, attrs=
			[Dot.Label (simplename (#name s)),
			 Dot.Shape Dot.Box]}
	in case #parent s of NONE => n :: k
		     | SOME p => (genname p --> this) [] ::n ::k
	end


fun compile1 (sys :(aor,atr) system) :codgen =
	let val globals = [Dot.NodeAttr [Dot.FontName "Helvetica",
						Dot.FontSize 10]]
	    val body = hfoldland fo fa [] (gettree (#hierarchy sys))
	    val dag = {id=S.name (#name sys), cluster=false, 
			stmts=globals @ body}
	in cg (Dot.print [(false,true,dag)]) end

fun compile (ofile :string option) (ifile :string) (m :(aor,atr) model) :unit =
	let val fbase = case ofile of NONE => Path.file ifile
				    | SOME f => f
	    val os    = TextIO.openOut ( fbase ^ ".dot")
	in 
		foldl (fn (s,os) => os <- compile1 s << "\n\n") os m;

		TextIO.closeOut os
	end

	

end


