(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenC1_back.sml,v
   Comment: Back-end of CodGenC1. Translates intermediate representation to C.
   Release: scope-0-16_final 

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   Does not actually dump the code, but builds everything what is needed in
   csystem record. Then a dumper actually solves the puzzle building real C
   files from snippets of code created here.

   CodGenC1_back.sml,v 1.83 2003/07/22 22:00:48 wasowski Exp 
*)

structure CodGenC1_back = struct 

(* some interface *)

open Err CodGenC1_LR CodGenC_back_types CodGenC_back Say

infixr 1 $ 
infix  4 INIT
infix  3 --

structure SE = SymbolEnv
structure S  = Symbol

(* definitions are static and could be put in some static file *)

val markers =      DEFINE "STMRK"   (ig 0) 
		-- DEFINE "HMRK"    (ig 1) 
		-- DEFINE "IMRK"    (ig 2)
		-- DEFINE "FPSCOP"  (ig 3)
		-- DEFINE "NFPSCOP" (ig 4)
		-- DEFINE "MARKNO"  (ig 5)


(* names of mnemonics used to annotate array cells *)

val lrmnemonics = DEFINE "HIST" eoc 
			-- DEFINE "EVNT"	(sg "\t\t/* event reference */")
			-- DEFINE "ENEX(n)"	(sg "\t\t/* entry/exit ref  */")
			-- DEFINE "ACGD(n)"     (sg "\t\t/* action/guard    */")
			-- DEFINE "PCNC(n)"	(sg "\t\t/* pos/neg-cond    */")
			-- DEFINE "ORST"	(sg "\t\t/* or-state ref    */")
			-- DEFINE "ANDST"	(sg "\t\t/* and-state ref   */")

(* add named LR mnemonic to compund (encoded) value *)

fun compound_mnemonic (name :string) (ints :int list) :string list =
	if null ints then []
	else String.concat [name, "(", itoa (length ints), ") ", itoa (hd ints)]
		::mitoa (tl ints)

(* pack entry/exit actions. NOTE: this must be in synchro with corresponding
   measure_enex function in scheduler *)

fun pack_enex (sizes :lrsizes) (en :int,ex :int) :string list =
	let val ints = 
		if ex = null_actref then
			if en = null_actref then []	
			else pack1 en (#actionref sizes) (#orstref sizes)
		else pack2 (ex,en) (#actionref sizes,#actionref sizes) 
			(#orstref sizes)
	in compound_mnemonic "ENEX" ints end



fun cg_getexit (sizes :lrsizes) :codgen =
	let val _ = assert (fn () => (#actionref sizes <= #orstref sizes))
			"200210211 - unhandled type sizes combination"
			(* see comment below for acgd *)
	in	
		if (#actionref sizes < #orstref sizes) then
			DEFINE "GET_EXIT(s)"
				(cg ["(((GET_AND(s)-2) == STMRK) ?",
					"((*(GET_AND(s)-1)) >> ",
					itoa (#actionref sizes), ")",
				     ": STMRK)"])
				(* if there is no action STMRK will do well!*)

		else (* #actionref siszes = #orstref sizes *)
			DEFINE "GET_EXIT(s)" (cg ["((*(GET_AND(s)-3) == STMRK ) ?",
						   "*(GET_AND(s)-2) : STMRK)"])				
	end

fun cg_getenter (sizes :lrsizes) :codgen = 
	let val _ = assert (fn () => (#actionref sizes <= #orstref sizes))
			"200210212 - unhandled type sizes combination"
			(* see comment below for acgd *)
	in	
		if (#actionref sizes < #orstref sizes) then
			DEFINE "GET_ENTER(s)"
				(cg ["*(GET_AND(s)-2) == STMRK ? ((*(GET_AND(s)-1)) & ",
				itoa (ones (#orstref sizes) + 1),
				") : STMRK "])
		else 
			DEFINE "GET_ENTER(s)" (cg ["((*(GET_AND(s)-2) == STMRK ||", 
						   "(*(GET_AND(s)-3) == STMRK )) ?",
						   "*(GET_AND(s)-1) : STMRK)"])
	end

(* pack action and guard fields - used in transitions. NOTE: should be in
   synchro with measure_acgd function of scheduler *)

fun pack_acgd (sizes :lrsizes) (ac :int, gd :int) :string list =
	let val ints = 
		if gd = always_ref then							
			if ac = null_actref then []
			else pack1 ac (#actionref sizes) (#andstref sizes)
		else pack2 (ac,gd) (#actionref sizes,#guardref sizes) 
			(#andstref sizes)
	in compound_mnemonic "ACGD" ints end

(* FIXME: there is a fundmanental problem with runtime representation *)
(* action and guard become indistinguishable from HMRK/IMRK as soon   *)
(* as sizeof(guardref) > sizeof(andstref) or                          *)
(* sizeof(actionref) > sizeof(andstref). I leave the problem unsolved *)
(* for a moment, hoping that this never holds for models I have. It   *)
(* should however be fixed one day. For instance: save the youngest   *)
(* cell first and then exclude all values with youngest cell equal to *)
(* HMRK / IMRK / STMRK from being legal guardref and actionref        *)
(* Alternatively change internal representation (only in this case)   *)
(* explicitly adding compulosory field with number of bytes for action*)
(* and guard in total <- for each transition                          *)
(* BAD: to make things worse some part of this logics is also kept in *)
(* fire () (in CodGenC1c)                                             *)
(* FOR NOW I LEAVE A SANITY ASSERTION HERE			      *)

fun cg_unmangle_acgd guards (sizes :lrsizes) :codgen =
	let val _ = assert (fn () => (#actionref sizes <= #andstref sizes)
			andalso (#guardref sizes <= #andstref sizes))
			"200210210 - unhandled type sizes combination"
			(* see comment above *)
	in if not guards then
		DEFINE "UNMANGLE_AC(ptr,i)"
		(sg "{ (i) = *(ptr); if ( (i) >= MARKNO ) (ptr)++; }")

	else if (#actionref sizes + #guardref sizes <= #andstref sizes) then

			DEFINE "UNMANGLE_ACGD(ptr,i,j)"
			(sg "{ " -- access2 (#actionref sizes,#guardref sizes)
				(#andstref sizes) -- sg "{if ((j)==0) j = MARKNO;} }")
			
	else 
		DEFINE "UNMANGLE_ACGD(ptr,i,j)"
		(sg "{ (i) = *ptr; \
			\if (i >= MARKNO) (ptr)++; \
			\(j) = *ptr; \
			\if ( (j) >= MARKNO) (ptr)++;}")

	end

(* compile a list of byte code entries. suitable to use in folds *)
(* FIXME: jakby to przerobic na codgen i sprawdzic czy predkosc  *)
(* sie zmienila?                                                 *)

fun comp_lr (sizes :lrsizes) ({hand_dict,hor_dict,eenv,...} :scheduling) 
	(atom :lratom,res :string list) : string list = (
	case atom of 
		  LRMrk mrk		=> comp_marker mrk :: res
		| LRAndStateRef s	=> "ANDST " ^ 
						itoa (lookup_stdict hand_dict s) 
						::res
		| LROrStateRef  s	=> ("ORST " ^ 
						itoa (lookup_stdict hor_dict  s) 
						::res 
						handle e => intErr e 200209259)
		| LRNoEvent		=> "NOEVENT" ::res
		| LREventRef e		=> "EVNT " ^ itoa (SE.lookup eenv e):: res
		| LRHistoryRef i	=> "HIST " ^ itoa i :: res
		| LREnEx (en,ex)	=> pack_enex sizes (en,ex) @ res
		| LRAcGd (ac,gd)	=> pack_acgd sizes (ac,gd) @ res
		| LRCondCounters(pc,nc)	=> pack_pcnc sizes (pc,nc) @ res
		| LRStateLabel _	=> res
		| LREventLabel _	=> res )
	handle e => intErr e 200209257

(* compile history state *)

fun cg_history (history :lratom list) (sizes :lrsizes) (sch :scheduling)
							:codgen =
	if history = [LRMrk STMRK] then 
		UNDEF "SCOPE_HISTORY"  (* no history saved *)
	else cg_arraylr "history" "andstref" (rev history) (comp_lr sizes sch)
		-- DEFINE "SCOPE_HISTORY" eoc

(* code generate access macro: macro which given stateid returns a pointer
   to the main cell of state representation *)

fun cg_stacc (dict :stdict) (dname :string) (hname :string) 
		(macro :string) :codgen =
	case dict of
		  Direct _	=> DEFINE (macro^"(d)") 
					(cg ["(",hname," + (d))"])
		| Indirect _	=> DEFINE (macro^"(d)") 
					(cg ["(",hname," + ", dname,"[d]"])

(* code-generate hierarchy access dictionaries if needed *)

fun cg_stdict (sizes :lrsizes) (dname :string) (dict :stdict) :codgen option =
	case dict of 
		  Direct _ => NONE (* no code generation needed *)
		| Indirect (mapping,lst,size) =>
			let val cty = "const " ^ (tyname $ uint_bysize 
							$ mintysize $ size-1)
			    val iniblock = mk_commented_arrini (mitoa lst) 10
			    val asize = itoa (length lst)
			in SOME (ARRAY_DECL cty dname asize INIT iniblock) end

(* always return the macro for accessing states          *)

fun cg_getst (dict :stdict) (macroname, arrayname, dictname) :codgen =
	case dict of
		  Direct _ => DEFINE (macroname ^ "(s)") 
			(cg ["(", arrayname, "+ (s))"])
		| Indirect _ => DEFINE (macroname ^ "(s)")
			(cg ["(", arrayname, "+ ", dictname, "[(s)] )"])

(* code generate symbolic dictionaries if needed *)

fun cg_symdict (debug :bool) (dname :string) (dict :stdict) :codgen option =
	let val d    = case dict of Direct d => d | Indirect (d,_,_) => d
	in if debug then SOME (cg_intenv dname d) else NONE end


(* #defines for PURE_EXIT and IMPURE_EXIT *)

fun cg_exitmode {pure,impure} :codgen =
	(if pure then DEFINE "PURE_EXIT" eoc 
	 else UNDEF "PURE_EXIT" )
	-- (if impure then DEFINE "IMPURE_EXIT"  eoc
	    else UNDEF "IMPURE_EXIT" )

(* Code generate system *)

fun cg_system (debug,externhdr,confset) (sys :lrsystem) (sizes :lrsizes) 
					(sch :scheduling) : csystem =
	let val (vardefs,vardecls) = cg_vars (#venv sys) (#vartypes sys)
					handle e => intErr e 200209251
	    val (trans,tranlen)    = cg_trans (comp_lr sizes sch) (#tenv sys)
	    val (tranidx,tranref)  = cg_tranidx (#tranidx sch) (#eenv sch) 
					sizes tranlen
					handle e => intErr e 200209253
	    val (guards,scopeguards)= cg_guards markno (#guards sys)
	    val (signals,scopeslist)= cg_signals (#signals sys) 
					(comp_lr sizes sch)
	    val unused = sg "    /*** unused in CodGenC1 ***/"
	in { name        = #name sys,
	     fname       = #fname sys,
	     codgen      = "CodGenC1",
	     lrmnemonics = lrmnemonics,

	     hor         = cg_arraylr "hor"  "const horcell"  (#hor sys)  (comp_lr sizes sch),
	     hand        = cg_arraylr "hand" "const handcell" (#hand sys) (comp_lr sizes sch),
	     hor_dict    = cg_stdict sizes "hor_dict"    (#hor_dict sch),
	     hand_dict   = cg_stdict sizes "hand_dict" (#hand_dict sch),
	     hor_acc     = cg_stacc (#hor_dict sch) "hor_dict" "hor" "ORST",
	     hand_acc    = cg_stacc (#hand_dict sch) "hand_dict" "hand" "ANDST",
	     and_syms    = cg_symdict debug "and_syms" (#hand_dict sch),
	     or_syms     = cg_symdict debug "or_syms"  (#hor_dict  sch),

	     symbol      = if debug then SOME struct_symbol else NONE,

	     trans       = trans,
	     init        = unused,
	     tranidx     = tranidx,
	     
	     evmacros    = cg_evmacros (#eenv sch) (#evtypes sys),
	     evdict      = cg_evdict debug (#eenv sch),
	     evtypes     = cg_evtypes (#evtypes sys),
	     evunion     = cg_evunion (#evtypes sys),
	     
	     vardefs     = vardefs,
	     vardecls    = vardecls,
	     consts      = cg_consts (#cenv sys),
	     
	     actions     = cg_actions markno (#eenv sch) (#evnumber sizes) 
					(#signals sys) (#actions sys),
	     guards      = guards,
	     externhdr   = cg_externhdr (#name sys) externhdr,
	     signals     = signals,
	     history     = cg_history (#history sys) sizes sch,

	     ROOT_STATE  = DEFINE "ROOT_STATE"  (sg $ itoa $ 
				lookup_stdict(#hand_dict sch) $ #root sys),
	     STATE_WIDTH = DEFINE "STATE_WIDTH" (sg $ itoa $ #statew sys),
	     TGT_WIDTH   = DEFINE "TGT_WIDTH"   (sg $ itoa $ #targetw sys),
	     EVENTNO     = DEFINE "EVENTNO"     (sg $ itoa $ #evnumber sizes)
				-- DEFINE "NOEVENT" (sg "EVENTNO"),
	     queue       = cg_queue (#queuelen sys),
	     markers     = markers,

	     UNMANGLE_PCNC
			 = cg_unmangle_pcnc sizes,
	     UNMANGLE_ACGD
			 = cg_unmangle_acgd (UList.length (#guards sys) > 1) sizes,
	     GET_OR	 = cg_getst (#hor_dict sch) ("GET_OR","hor","hor_dict"),
	     GET_AND	 = cg_getst (#hand_dict sch)("GET_AND","hand","hand_dict"),
	     GET_ENTER   = cg_getenter sizes,
	     GET_EXIT    = cg_getexit sizes,

	     SCOPE_GUARDS= scopeguards,
	     SCOPE_SLIST = scopeslist,
	     SCOPE_CONFSET
			 = if confset then DEFINE "SCOPE_CONFSET" eoc
			   else UNDEF "SCOPE_CONFSET",

	     exitmode    = cg_exitmode (#exitmode sys),

	     prev_conf   = cg_conf "prev_conf" (#statew sys),
	     next_conf   = cg_conf "next_conf" (#statew sys),
	     targets     = cg_conf "targets"   (#targetw sys),
	     
	     actionref   = cg_typedef "actionref" (#actionref sizes),
	     guardref    = cg_typedef "guardref"  (#guardref  sizes),
	     orstref     = cg_typedef "orstref"   (#orstref   sizes),
	     andstref    = cg_typedef "andstref"  (#andstref  sizes),
	     signalref   = cg_typedef "signalref" (#signalref sizes),
	     eventtag    = cg_typedef "eventtag"  (mintysize $ #evnumber sizes),
	     traniter    = cg_typedef "traniter"  (mintysize $ tranlen),
	     tranref     = tranref,

	     confiter    = cg_typedef "confiter"  (mintysize $ #statew sys),
	     conf        = TYPEDEF "andstref" ("conf["^(itoa $ #statew sys)^"]"),

	     tgtsiter    = cg_typedef "tgtsiter"  (mintysize $ #targetw sys),
	     queueiter   = case #queuelen sys of 
				  NONE => NONE
				| SOME n => SOME(cg_typedef "queueiter" (mintysize n)),
	     
	     (* model-independent typedefs, included here to avoid *)
	     (* creating more static files                         *)

	     types       = TYPEDEF "andstref" "horcell" -- eol
				-- TYPEDEF "orstref"  "handcell" -- eol
				-- TYPEDEF "andstref" "transcell" -- eol 
				-- TYPEDEF "andstref" "historyref" -- eol
	     }
	end handle e => intErr e 200209254

end
