/* File : /home/cvs/cvsroot/projects/visualstate/scope/CodGenCF.c,v 
 * Comment: Runtime interpreter for programs generated with CodGenCF
 * Release: scope-0-16_final
 * 
 * Copyright (C) 2003 Andrzej Wasowski
 * 
 * This program is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along 
 * with this program; if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * CodGenCF.c,v 1.11 2004/01/26 12:11:58 wasowski Exp 
 */

/**
 * a hack to compile an avr version easily 
 * (unfortunately no assert in avr lib, at least in the one I got installed)
 */

#ifdef NDEBUG
#define assert(p)
#else
#include <assert.h>
#endif

#define SCOPE_VERBOSE 0

#if (SCOPE_VERBOSE && !defined(NDEBUG))
	#define VERBOSE(msg)		{ printf ( "%s", msg ); }
#else
	#define VERBOSE(msg)
#endif

#ifndef NDEBUG

#include <stdlib.h>
#include <stdio.h>

#if STATE_WIDTH > 0
/* check if conf is empty (used in assertions) */
static int conf_empty(conf c)
{
	int i = 0;
	for (; i < STATE_WIDTH; i++)
		if (c[i] != STMRK)
			return 0;
	return 1;
}

/* check if state s is a member of c */
int conf_member_ext(conf c, stref s)
{
	return c[anatomy[s]] == s;
}

void next_conf_reset(void)
{
	confiter i;
	for (i = 0; i < STATE_WIDTH; ++i)
		next_conf[i] = STMRK;
}

int compar(const void *a, const void *b)
{
	if (*(stref *) a < *(stref*) b)
		return -1;
	else if (*(stref *) a == *(stref *) b)
		return 0;
	return 1;
}

void print_st_sym(stref s)
{
	int c = 0;
	while (1) {
		if (st_syms[c].id == s) {
			printf("%s ", st_syms[c].name);
			break;
		} else ++c;
	}
}

void print_mchn_sym(mchnref s)
{
	int c = 0;
	while (1) {
		if (mchn_syms[c].id == s) {
			printf( "%s", mchn_syms[c].name);
			break;
		} else ++c;
	}
}

/* check if what is a state. Only works in --debug mode, bcs it uses
 * symbolic dictionaries 
 */

unsigned is_state(stref what)
{
	int c = 0;
	while (st_syms[c].id != 0) {
		if (st_syms[c].id == what)
			return 1;
		c++;
	}
	return 0;
}

/*
 * check if what is an orstate. Only works in --debug mode, bcs it uses
 * symbolic dictionaries 
 */

unsigned is_mchn(mchnref what)
{
	int c = 0;
	while (mchn_syms[c].id != 0) {
		if (mchn_syms[c].id == what)
			return 1;
		c++;
	}
	return 0;
}
#endif /* STATE_WIDTH > 0 */

void dump_conf()
{
#if STATE_WIDTH > 0
	int i = 0;
	conf myconf;

	for (i = 0; i < STATE_WIDTH; ++i)
		myconf[i] = next_conf[i];

	printf("conf: [");
	qsort(myconf, STATE_WIDTH, sizeof(stref), compar);
	for (i = 0; i < STATE_WIDTH; ++i) 
		if (myconf[i] != STMRK)
			print_st_sym(myconf[i]);
	printf("]\n");
#else	
	printf("[No discrete states in this model]\n");
#endif
}

#if STATE_WIDTH > 0
VS_BOOL activest_next(const stref what)
{
	assert(is_state (what));
	assert(anatomy[what] < STATE_WIDTH);
	return next_conf[anatomy[what]] == what;
}
#endif

#endif	/* NDEBUG */



#if STATE_WIDTH > 0
VS_BOOL active(const stref what)
{
	return prev_conf[anatomy[what]] == what;
}
#endif

#ifdef SCOPE_SIGNALQUEUE

/* Signal Queue - implemented as a round buffer */
static queueiter qbeg = 0;
static queueiter qend = 0;

/*
 * silently assume that queue is long enough! which shall be the case,
 * since verification returns this result 
 */
void enqueue1(eventtag e)
{

	assert(e < EVENTNO);

	queue[qend] = e;

	if (qend == QUEUE_LENGTH - 1)
		qend = 0;
	else
		qend++;

	assert(qend != qbeg);
}

#ifdef SCOPE_SLIST
void enqueueList(signalref ref)
{
	while (signals[ref] != NOEVENT)
		enqueue1(signals[ref++]);
}
#endif

/*
 * returning NOEVENT means that the queue is empty result is returned in
 * CurrEvent.tag (global variable) That is the way you do it in this funny 
 * language to conserve space ... 
 */
void dequeue1(void)
{
	if (qend != qbeg) {	/* non empty */

		CurrEvent.tag = queue[qbeg];
		if (qbeg == QUEUE_LENGTH - 1)
			qbeg = 0;
		else
			qbeg++;
	} else
		CurrEvent.tag = NOEVENT;
}

#endif	/* SCOPE_SIGNALQUEUE */




/*************************************
 * Implementation of The Interpreter *
 *************************************/



/* Fire a transition. Transition is assumed to be enabled.  The parameter
 * passed is assumed to be the index in transitions array were the
 * guard/action indicators are kept. We point into the middle of the
 * record for slight efficiency gain. The preceding part will not be used
 * in this function anyway. The assumption is that targets are empty on
 * entry and on exit. 
 */

void fire(const transcell * ptr)
{
	actionref ac = STMRK;

#ifdef SCOPE_GUARDS
	guardref gd = STMRK;

	UNMANGLE_ACGD(ptr, ac, gd);
	if (!eval(gd))
		return;
#else
	UNMANGLE_AC(ptr, ac);
#endif
	VERBOSE("TRANAC:")
	exec(ac);
	VERBOSE(" TRANAC\n")
#if STATE_WIDTH > 0
	while (*ptr != STMRK) {
		assert(is_state (*ptr));
		next_conf[anatomy[*ptr]]=*(ptr++);	
		
	}
#else
	assert(*ptr == STMRK); /* no state machines */
#endif
}

/* microstep - assumes that proper event is already in CurrEvent no
 * conflict resolution currently. Everything will collapse on conflict 
 */

void microstep(void)
{

	const transcell *t = trans + tranidx[CurrEvent.tag];

	assert(CurrEvent.tag < NOEVENT);

	if (tranidx[CurrEvent.tag] != TRANS_MAX)
		do {

			unsigned int pc, nc;

			UNMANGLE_PCNC(t, pc, nc);
#if STATE_WIDTH > 0
			for (; pc > 0; --pc, ++t)
				if (!active(*t))
					goto skiptr;

			for (; nc > 0; --nc, ++t)
				if (active(*t))
					goto skiptr;
#else
			assert (pc==0 && nc==0); /* they could be left out... */
#endif

			fire(t);	/* checking guard is deferred to fire */
#if STATE_WIDTH > 0
		      skiptr:while (*t != STMRK)
				++t;
#else
			if (*t != STMRK) ++t;
			assert(*t == STMRK);
#endif

			++t;

		} while (t < trans + tranidx[CurrEvent.tag + 1]);
}

/* assumes that external event is already in CurrEvent the newest
 * configuration is always kept in next_conf. prev_conf is only occasionally
 * used for guard computations. 
 */
void macrostep(void)
{

#ifdef SCOPE_SIGNALQUEUE
	while (CurrEvent.tag != NOEVENT) {
#endif
#if STATE_WIDTH > 0
		confiter i = 0;
		for (; i < STATE_WIDTH; ++i)
			prev_conf[i] = next_conf[i];

		assert(!conf_empty(next_conf));
#endif
		
		microstep();

#ifdef SCOPE_SIGNALQUEUE
		dequeue1();
	}
#endif

}

