(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenCF_back.sml,v
   Comment: Back-end of CodGenCF. Translates intermediate representation to C.
   Release: scope-0-16_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   Does not actually dump the code, but builds everything what is needed in
   csystem record. Then a dumper actually solves the puzzle building real C
   files from snippets of code created here.

   CodGenCF_back.sml,v 1.17 2003/07/27 12:39:29 wasowski Exp 
*)

structure CodGenCF_back = struct 

(* some interface *)

open Err CodGenCF_LR CodGenC_back_types CodGenC_back Say

infix  4 INIT
infix  3 --

structure SE = SymbolEnv
structure S  = Symbol

val lrmnemonics =  DEFINE "EVNT"	(sg "\t\t/* event reference */")
		-- DEFINE "ACGD(n)"     (sg "\t\t/* action/guard    */")
		-- DEFINE "AC(n)"	(sg "\t\t/* action(no guard)*/")
		-- DEFINE "PCNC(n)"	(sg "\t\t/* pos/neg-cond    */")
		-- DEFINE "MCHN"	(sg "\t\t/* machine ref     */")
		-- DEFINE "STATE"	(sg "\t\t/* basic state ref */")

(* pack action and guard fields - used in transitions. NOTE: should be in
   synchro with measure_acgd function of scheduler. This is simpler than
   in C1 as actions and guards are *always* present *)

fun pack_acgd (guards :bool) (sizes :lrsizes) (ac :int, gd :int) :string list =
	if guards then 
		let val ints = pack2 (ac,gd) (#actionref sizes,#guardref sizes) 
				(#andstref sizes)
		in compound_mnemonic "ACGD" ints end
	else compound_mnemonic "AC" (pack1 ac (#actionref sizes) (#andstref sizes))


fun cg_unmangle_acgd (hasguards :bool) (sizes :lrsizes) :codgen =
	if not hasguards then
		DEFINE "UNMANGLE_AC(ptr,i)" (sg "{ (i) = *(ptr++); }")
	else if (#actionref sizes + #guardref sizes <= #andstref sizes) then

			DEFINE "UNMANGLE_ACGD(ptr,i,j) " 
			(sg "{ " -- access2 (#actionref sizes,#guardref sizes)
				(#andstref sizes) -- sg " }")
	else 
		DEFINE "UNMANGLE_ACGD(ptr,i,j) " 
			(sg "{ (i) = *(ptr++); (j) = *(ptr++);}")



(* compile a list of byte code entries. suitable to use in folds *)
(* FIXME: jakby to przerobic na codgen i sprawdzic czy predkosc  *)
(* sie zmienila?                                                 *)

fun comp_lr (hasguards :bool) (sizes :lrsizes) ({st_dict,m_dict,eenv,...} :scheduling) 
	(atom :lratom,res :string list) : string list = (
	case atom of 
		  LRMrk mrk		=> comp_marker mrk :: res
		| LRAndStateRef s	=> "STATE " ^ itoa (SE.lookup st_dict s)
						::res
		| LROrStateRef  s	=> ("MCHN " ^ itoa (SE.lookup m_dict s) 
						::res 
						handle e => intErr e 200307200)
		| LRNoEvent		=> "NOEVENT" ::res
		| LREventRef e		=> "EVNT " ^ itoa (SE.lookup eenv e):: res
		| LRHistoryRef _	=> impossible "#200307201"
		| LREnEx (en,ex)	=> impossible "#200307202"
		| LRAcGd (ac,gd)	=> pack_acgd hasguards sizes (ac,gd) @ res
		| LRCondCounters(pc,nc)	=> pack_pcnc sizes (pc,nc) @ res
		| LRStateLabel _	=> res
		| LREventLabel _	=> res )
	handle e => intErr e 200307219

(* initial transition *)

fun cg_init (sch :scheduling) (sigs: signals) comp_lr (init :action * lratom list) = 
	let
	    val iniconf = cg_arraylr "iniconf" "static const stref" (#2 init) comp_lr
	    val copy    = cg ["\tfor (i=0; i < STATE_WIDTH; i++)\n",
				"\t\tnext_conf[i]=iniconf[i];\n\n"]
	    val abody   = cg_cmds (#1 (#1 init))
	    val sbody   = case (#2(#1 init)) of
				  []  => eoc
				| [s] => sg "enqueue1( " -- ig (SE.lookup 
						(#eenv sch) (#1 s)) 
						-- sg " );\n"
				| ss  => sg "enqueueList( " 
						-- ig (signals2int sigs ss) 
						-- sg " );\n"
	    val body    = sg "\n#if STATE_WIDTH > 0\n"
				-- VAR_DECL "confiter" "i = 0;\n\n"
				-- iniconf -- copy --
			  sg "#endif\n" 
				-- abody -- sbody --
			  sg "\n#ifdef SCOPE_SIGNALQUEUE\n \
				      \\tdequeue1();\n\tmacrostep();\n\
			      \#endif\n"
	in FUNC (sg "void st_init(void)") body end
	handle e => intErr e 200307233
  

(* Code generate system *)

fun cg_system (debug,externhdr) (sys :lrsystem) (sizes :lrsizes) 
					(sch :scheduling) : csystem = 
	let val (guards,scopeguards)= cg_guards markno (#guards sys)
	    val hasguards          = UList.length (#guards sys) > 1
	    val (vardefs,vardecls) = cg_vars (#venv sys) (#vartypes sys)
	    val (trans,tranlen)    = cg_trans (comp_lr hasguards sizes sch) (#tenv sys)
	    val (tranidx,tranref)  = cg_tranidx (#tranidx sch) (#eenv sch) 
					sizes tranlen 
	    val (signals,scopeslist)= cg_signals (#signals sys) 
					(comp_lr hasguards sizes sch)
	    val unused  = sg "    /*** unused in CodGenCF ***/"
	    val comp_lr = comp_lr hasguards sizes sch
	    val hasconf = #statew sys > 0
	in { name         = #name sys,
	     fname        = #fname sys,
	     codgen       = "CodGenCF",
	     lrmnemonics  = lrmnemonics,

	     hor          = eoc, 
	     hand         = cg_arraylr "anatomy" "const anatomycell" 
				(#anatomy sys) comp_lr,

	     hor_dict     = NONE,
	     hand_dict    = NONE, 
	     hor_acc      = unused,
	     hand_acc     = unused, 
	     and_syms     = if not debug then NONE
			    else SOME (cg_intenv "st_syms" (#st_dict sch)),
	     or_syms      = if not debug then NONE
			    else SOME (cg_intenv "mchn_syms"  (#m_dict  sch)),

	     symbol       = if debug then SOME struct_symbol else NONE,

	     trans        = trans,
	     init         = cg_init sch (#signals sys) comp_lr (#init sys), 
	     tranidx      = tranidx,
	     
	     evmacros     = cg_evmacros (#eenv sch) (#evtypes sys), 
	     evdict       = cg_evdict debug (#eenv sch),
	     evtypes      = cg_evtypes (#evtypes sys),
	     evunion      = cg_evunion (#evtypes sys),
	     
	     vardefs      = vardefs,
	     vardecls     = vardecls,
	     consts       = cg_consts (#cenv sys),
	     
	     actions      = cg_actions markno (#eenv sch) (#evnumber sizes) 
					(#signals sys) (#actions sys),
	     guards       = guards,
	     externhdr    = cg_externhdr (#name sys) externhdr,
	     signals      = signals,
	     history      = unused,

	     ROOT_STATE   = unused,
	     STATE_WIDTH  = DEFINE "STATE_WIDTH" (sg (itoa (#statew sys))),
	     TGT_WIDTH    = unused,
	     EVENTNO      = DEFINE "EVENTNO"     (sg (itoa (#evnumber sizes)))
				-- DEFINE "NOEVENT" (sg "EVENTNO"),
	     queue        = cg_queue (#queuelen sys), 
	     markers      = DEFINE "STMRK" (ig 0),

	     UNMANGLE_PCNC= cg_unmangle_pcnc sizes, 
	     UNMANGLE_ACGD= cg_unmangle_acgd hasguards sizes, 
	     GET_OR	  = unused, 
	     GET_AND	  = unused, 
	     GET_ENTER    = unused,
	     GET_EXIT     = unused,

	     SCOPE_GUARDS = scopeguards,
	     SCOPE_SLIST  = scopeslist,
	     SCOPE_CONFSET= unused,

	     exitmode     = unused,

	     prev_conf    = if hasconf then cg_conf "prev_conf" (#statew sys)
			    else unused,
	     next_conf    = if hasconf then cg_conf "next_conf" (#statew sys)
			    else unused,
	     targets      = unused, 
	     
	     actionref    = cg_typedef "actionref" (#actionref sizes),
	     guardref     = cg_typedef "guardref"  (#guardref  sizes),
	     orstref      = cg_typedef "mchnref"   (#orstref   sizes),
	     andstref     = cg_typedef "stref"     (#andstref  sizes),
	     signalref    = cg_typedef "signalref" (#signalref sizes),
	     eventtag     = cg_typedef "eventtag"  (mintysize (#evnumber sizes)),
	     traniter     = cg_typedef "traniter"  (mintysize tranlen), 
	     tranref      = tranref, 

	     confiter     = cg_typedef "confiter" (mintysize (#statew sys)),
	     conf         = if not hasconf then unused
			    else TYPEDEF"stref"("conf["^(itoa(#statew sys))^"]"),

	     tgtsiter     = unused,
	     queueiter    = case #queuelen sys of 
				  NONE => NONE
				| SOME n => SOME(cg_typedef "queueiter" (mintysize n)),
	     
	     (* model-independent typedefs, included here to avoid *)
	     (* creating more static files                         *)

	     types        = TYPEDEF "mchnref"  "anatomycell" -- eol
				   -- TYPEDEF "stref" "transcell" -- eol 
	     }
	end handle e => intErr e 200307218 

end
