(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenCF_schedule.sml,v
   Comment: Scheduler for CodGenCF. The name scheduler is used a bit
            improperly. See note below.
   Release: scope-0-16_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   The task of the scheduler is to resolve all string references in
   intermediate representation code generated by CodGenCF. Scheduler takes
   lrsystem and returns scheduling data for it.
   
   CodGenCF_schedule.sml,v 1.4 2003/07/23 11:51:15 wasowski Exp *)

structure CodGenCF_schedule = struct

(* some interface *)

open Err Say CodGenCF_LR CodGenC_schedule

(* measure size of bulk of transitions *)

fun measure_trans (sizes :lrsizes) hasguards (arr :lratom list)  :int =
	let val trcell = #andstref sizes
	    val pcnc   = if (trcell >= 2) then 1 else 2
	    val acgd   = if hasguards then measure2 
				(#actionref sizes, #guardref sizes) trcell
			 else measure1 (#actionref sizes) trcell
	    fun f (LRMrk          _,e) = e + 1
	      | f (LRAndStateRef  _,e) = e + 1
	      | f (LRAcGd   (ac,gd),e) = e + acgd
	      | f (LRCondCounters _,e) = e + pcnc 
	      | f (LRStateLabel   _,e) = e + 0 
	      | f (LREventLabel   _,e) = e + 0 
	      | f _  = impossible "#200307216: unexpected LR opcode in trans"
	in foldl f 0 arr end


(* build a dictionary for accessing states       *)
(* the same function for or-state and and-states *)
(* offset is the number of cells to reserve in   *)
(* the begining of dictionary to avoid clashes   *)
(* with markers of LR                            *)


fun build_dict (arr :lratom list) :int env =
	let fun f lst n max dict = case lst of 
			  [] => (max, dict)
			| LRStateLabel s::ta => f ta n n (SE.bind1 dict (s,n))
			| atom::ta => f ta (n+1) max dict
	    val (max,dict) = f arr 0 0 SE.empty
	in dict end

(* resolve or labels, assign addresses and compile to int lists *)

fun schedule_system (sys :lrsystem) (sizes :lrsizes) :scheduling =
	let val measure_tr = measure_trans sizes 
				(UList.length (#guards sys) > 1)
	in { m_dict    = build_dict (#mconf sys),
	     st_dict   = build_dict (#anatomy sys),
	     eenv      = schedule_events (#evtypes sys),
	     tranidx   = build_tranidx measure_tr (#tenv sys) (#evtypes sys) } 
	end
	
		
end
