(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/CodGenDot.sml,v
   Comment: Draws a staechart using Dot
   Release: scope-0-16_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   CodGenDot.sml,v 1.7 2003/07/08 15:45:01 wasowski Exp
*)

structure CodGenDot : CODGEN = struct

open Absyn CodGen Cppish
infix 0 << <<< <<> <- (* some fun with C++ style IO *)
infix 3 --

val --> = Dot.-->
infix 0 --> (* edge constructor for Dot *)

structure S = Symbol
structure D = Dot
structure AP = AbsynPrint

type options = string option
type aor = unit
type atr = unit


fun simplename (s :S.symbol) :string=
	List.last (String.tokens (fn c => c = #".") (S.name s))

fun genname (s :S.symbol) :string =
	"node" ^ Int.toString (S.toInt s)

fun faction prefix a :string =
	if a = ([],[]) then ""
	else String.concat ["\\n", prefix, " / ", AP.sprAction a]

fun ininame (s :S.symbol) :S.symbol = S.symbol ("INITIAL:" ^ S.name s)

fun fini H (ANDState s) :D.stmt list =
	let val par   = valOf (#parent s) 
			handle Option.Option => S.symbol "_TOP"
	    val iniid = genname (ininame par)
	    val nameid= genname (#name s)
	    val (attrs,head) = if null (#children s) then ([],nameid)
			       else ([D.LHead nameid,D.Label "OLA"],
				     (genname o ininame o #name) s)
	in	
		D.Node {id=iniid, attrs=[D.Shape D.Circle, D.Label H,
					 D.Height 0.16, D.Style D.Filled,
					 D.FillColor D.Black,
					 D.FontColor D.White,
					 D.FontSize 9,
					 D.FixedSize true]}
			:: (iniid --> head) (attrs) :: []
	end

fun fo single (ORState s) :D.stmt =
	let val this  = genname (#name s)
	    val H     = if #marking s = History then "H" else ""
	    val stmts = D.GraphAttr [D.Label (simplename(#name s)),
				     D.Style D.Dashed] 
			 :: fini H (hd (#children s))
			 @ map fa (#children s) 
	in D.Subgraph {id=this, cluster=true, stmts=stmts} end

and fa (ANDState s) :D.stmt =
	let val this = genname (#name s)
	    val n    = simplename (#name s)
	    val act  = faction "Entry" (#entry s) ^
		       faction "Exit"  (#exit s)
	in if null (#children s) then (* BASIC *) 
		Dot.Node {id=this, attrs= [D.Label (n^act),
					   D.FillColor D.LightGray,
					   D.Style D.Filled]}
	   else
		let val desc = map (fo (length (#children s)=1)) 
					(#children s)
		    val stmts = D.GraphAttr 
					[Dot.Label (n^act),
					 D.Style D.Solid]
				:: D.Node {id=genname (ininame (#name s)),
					   attrs=[D.Shape D.Point,
						  D.Height 0.0,
						  D.Color D.White ] }
				:: desc
		in D.Subgraph {id=this, cluster=true, stmts=stmts} end
	end

fun ft h ({trigger,pos, guard, targets,action,...} :atr tran) :D.stmt list =
	let val tail = genname (hd pos)
	    val head = genname (hd targets)
	    val stri = AP.sprTrigger trigger 
	    val sgua = if null guard then ""
		       else "\\n[" ^ AP.sprGuard guard ^ "]"
	    val sact = if action = ([],[]) then ""
		       else " / \\n" ^ AP.sprAction action
	    val src  = getandh h (hd pos)
	    val tgt  = getandh h (hd targets)
	    val attrs= [D.Label (stri ^ sgua ^ sact)]

	    (* a bit complicated algorithm introducing a dummy *)
	    (* intermediate point if transition has target     *)
	    (* and scope ancestrally related                   *)

	    val (attrs,tail) = 
			if not (isbasicand src) then 
				(D.LTail tail::attrs,               
				 (genname o ininame o hd) pos)      
			else (attrs,tail)                              
	    val (attrs,tail,stmts) = 
			if ancestralh h (hd pos) (hd targets) then
				let val fresh = genname (S.fresh())
				    val attrs = D.ArrowHead D.None
							::attrs
				    val node  = D.Node {
					   id=fresh,
					   attrs=[D.Shape D.Point,
						  D.Height 0.0,
						  D.Color D.Black ] }
				in ([],fresh,
				    [(tail --> fresh) attrs,node]) 
				end
			else (attrs,tail,[])
	    val (attrs,head) = 
			if not (isbasicand tgt) then
				(D.LHead head::attrs,
				 (genname o ininame o hd) targets)
			else (attrs,head)
	in rev ((tail --> head) attrs :: stmts) end

fun compile1 (sys :(aor,atr) system) :codgen =
	let val body = [D.GraphAttr [D.FontName "Helvetica",
				     D.FontSize 10, D.Compound true],
			D.NodeAttr  [D.FontName "Helvetica",
				     D.FontSize 10, D.Shape D.Box],
			D.EdgeAttr  [D.FontName "Helvetica",
				     D.FontSize 8,
				     D.LabelFontName "Helvetica"],
			fa (gettree (#hierarchy sys))]
	    val trans = List.concat (map (ft (#hierarchy sys))
				    (#trans sys))
	    val dag = {id=S.name (#name sys), cluster=false, 
		       stmts=body @ trans}
	in cg (Dot.print [(false,true,dag)]) end

fun compile (ofile :string option) (ifile :string) m :unit =
	let val fname = case ofile of NONE => Path.file ifile ^ ".dot"
				    | SOME f => f
	    val os    = TextIO.openOut fname
	in 
		foldl (fn (s,os) => os <- compile1 s << "\n\n") os m;

		TextIO.closeOut os
	end

	

end


