(* -*- sml-lex -*-
   File   : /home/cvs/cvsroot/projects/visualstate/scope/envspec/envsetlex.lex,v
   Comment: Lexer for "set-detecting" environments
   Release: scope-0-16_final

     Copyright (C) 2003 Andrzej Wasowski

     Some portions from public materials from PFOO course:
     Copyright (C) 2001 Peter Sestoft 
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   envsetlex.lex,v 1.2 2003/12/01 15:48:47 wasowski Exp
*)
{
  open Lexing envsetpar;
  open StringCvt;

  (* error handling primitives *)
  
  exception LexicalError of string * int * int;

  fun mkLoc lexbuf = Location.Loc(getLexemeStart lexbuf, getLexemeEnd lexbuf)
  fun lexerError lexbuf msg = 
  	raise LexicalError (msg,getLexemeStart lexbuf,getLexemeEnd lexbuf)

  val commentStart = ref 0;  (* Start of outermost comment being scanned *)

  fun commentNotClosed lexbuf =
  	lexerError lexbuf "Comment not terminated";

  val commentDepth = ref 0;  (* Current comment nesting *)

  fun keyword s lexbuf =
    case s of
        "val"     => T_val
      | "true"    => T_true
      | "false"   => T_false
      | "event"   => T_event
      | "if"      => T_if
      | "then"    => T_then
      | "else"    => T_else
      | "let"     => T_let
      | "in"      => T_in
      | "end"     => T_end
      | "action"  => T_action
      | name      => T_id name

}                                 

rule Token = parse
   [` ` `\t` `\n` `\r`] { Token lexbuf }
|  [`a`-`z``A`-`Z`][`a`-`z``A`-`Z``0`-`9``_`]*[`'`]*
			{ keyword (getLexeme lexbuf) lexbuf }
| "(*"                  { commentStart := getLexemeStart lexbuf;
                          commentDepth := 1;
                          SkipComment lexbuf; Token lexbuf }
| "&&"			{ T_and      } 
| "||"                  { T_or       }
| "->"                  { T_imp      }
| "<->"                 { T_equiv    }
| "="                   { T_equ      }
| "!="                  { T_cneq     }
| "<>"                  { T_neq      }
| "+"                   { T_add      }
| "."                   { T_dot      }
| "!"                   { T_not      }
| "("                   { T_lpar     }
| ")"                   { T_rpar     }
| "{"                   { T_lbr      }
| "}"                   { T_rbr      }
| ";"                   { T_semi     }
| ","                   { T_comma    }
| eof                   { T_eof }
| _                     { lexerError lexbuf "Illegal symbol in input" }

and SkipComment = parse
    "*)"                { commentDepth := !commentDepth - 1;
                          if !commentDepth = 0 then ()
                          else SkipComment lexbuf
                        }
   | "(*"               { commentDepth := !commentDepth + 1;
                          SkipComment lexbuf }
   | (eof | `\^Z`)      { commentNotClosed lexbuf }
   | _                  { SkipComment lexbuf }
;
