(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/flat.sml,v
   Comment: Representation of flat models
   Release: scope-0-16_final

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   flat.sml,v 1.6 2003/07/22 21:32:37 wasowski Exp
*)

structure Flat = struct

open Common

type machineid = stateid	(* name of the machine taken from respective or-state *)
type anatomy   = machineid env  (* maps state names to machine names *)

(* type of transitions in the flat system *)

type tran = { trigger :trigger,
	      pos     :stateid list,		(* always basic-states *)
	      neg     :stateid list,		(* always basic-states *)
	      guard   :guard,
	      action  :action,
	      targets :stateid list }		(* always basic-states *)

(* some setters for tran *)

fun setcond (t :tran) (p,n) :tran =
	{ trigger= #trigger t, pos= p, neg= n, guard= #guard t, 
	  action= #action t, targets= #targets t }

(* type of flat systems *)

type system = {

	name     : symbol,
	fname    : symbol,

	machines : machineid list,  (* just lists machine names for efficiency *)
	anatomy  : anatomy,         (* maps basic-state names to machine names *)

	vartypes : vartype env,
	evtypes  : evtype  env,
	evgroups : evgroup env,
	otypes   : otype   env,

	denv     : stateid env,
	venv     : venv,
	cenv     : const env,

	init     : action * stateid list,	(* init action, targets list *)
	trans    : tran list
}

type model = system list


end

