(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/util/env-fn.sml,v 
   Comment: General efficient environments inspired by Andrew Appel's "Modern
            Compiler Implementation". Interface to environments inspired by
	    Peter Sestoft in PFOO course at IT-Universty, Spring 2002. 
   Release: scope-0-16_final

     Copyright (C) 2001-2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   NOTES:

   The interface is similar to one proposed by Peter. Only key is no longer a
   type variable (fo equality type), but a fixed type (so it is dropped from
   many parameters). key may be any time for which toInt function may be
   given to the functor. 

   Structure HashEnv implements environments from Symbol.symbol to any type.
   See file HashEnv.sml

   env-fn.sml,v 1.8 2003/07/07 19:02:19 wasowski Exp
*)   

functor EnvFn (s : sig type key
                       val  toInt: key -> int
		   end ) : ENV where type key = s.key =
struct

  structure I = Intmap
  type    key = s.key

  exception NotFound = I.NotFound

  type 'a env = (key * 'a) I.intmap
  
  val empty = I.empty()

  fun isEmpty env = I.numItems env <> 0
  
  val count = I.numItems

  fun lookup env k = 
	#2(I.retrieve(env, s.toInt k)) (* raises NotFound *)

  fun defined env k = case I.peek (env, s.toInt k) of
				  NONE   => false
				| SOME _ => true

  fun bind1 env (k,d) = 
	I.insert( env, s.toInt k, (k,d) )

  fun remove1 env k =
	#1 (I.remove( env, s.toInt k ))

  fun remove env ks =
	foldl (fn (k,env)=> remove1 env k) env ks

  fun bind env lst = 
	foldl (fn ((k,d),e) => I.insert(e,s.toInt k,(k,d))) env lst

  fun bindZip env lsts = 
	ListPair.foldl (fn (k,d,e) => I.insert(e,s.toInt k,(k,d))) env lsts

  fun fromList lst = 
	bind (I.empty()) lst

  fun toList env = 
	map #2 (I.listItems env)

  fun fullmap f  env = 
	I.map (fn (i,(k,d)) => (k,f (k,d))) env

  fun map f env = fullmap (fn (k,d) => f d) env

  fun plus (env1,env2) =  
	let val lst2 = I.listItems env2
	in foldl (fn ((k,d),e) => I.insert(e,k,d)) env1 lst2 end

  (* inefficient *)
  fun join (f :'a * 'a -> 'a) (env1 :'a env,env2 :'a env) =
	let val lst2 = I.listItems env2
	    fun g ((k,(sym,dat)),e) = 
			let val old = #2 (I.retrieve (e,k))
			in I.insert(e,k, (sym,f(old,dat))) end
			handle NotFound => I.insert(e,k,(sym,dat))
	in foldl g env1 lst2 end 

  fun minus (env,ks) =
	foldl (fn (k,env) => remove1 env k handle NotFound => env) env ks
  

  fun filter (p :key ->bool) (env :'data env) :'data env = 
	I.foldl (fn (i,(k,d),e) => 
			if p k then I.insert(e,i,(k,d)) else e) (I.empty()) env
end


