(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/wuddy/bddutil.sml,v
   Comment: Some utilities for BDDs implemented in Muddy
   Release: scope-0-16_final
     
     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

     bddutil.sml,v 1.1 2003/01/23 20:25:03 wasowski Exp
*)

structure BDDUtil : BDDUTIL = struct 

  open bdd

    (* existentially quantify over everything not in the given set ss *)

  fun restrict_to (r :bdd) (inter :int list) :bdd =
	let val universe = bdd.scanset (bdd.support r)
	    val noninter = Vector.foldl (fn (a,e) => 
				if List.all (fn v => v <> a) inter then a::e else e)
				[] universe
	in bdd.exist (bdd.makeset noninter) r end

    (* generalized n-argument xor : builds a formula
       which is true if exactly one of variables is true *)

    (* this is my quadratic algorithm *)

    fun generalized_xor' (vs :int list) :bdd =
	let val bdds = Vector.fromList (map ithvar vs)
	    fun mk i (j,b,f) :bdd =
			if j = i then AND (Vector.sub(bdds,j), f)
			else AND (NOT (Vector.sub(bdds,j)), f)
	    fun mk1 i :bdd = Vector.foldli (mk i) TRUE (bdds,0,NONE)
	    val ands  :bdd list = List.tabulate(length vs, mk1)
	in foldl OR FALSE ands end;

    (* Peter's proposal of nlog n which seems to work well, but still
       needs better performance evaluation. *)

    fun generalized_xor (vs :int list) :bdd = 
	let fun split ([] :int list) : bdd * bdd = (FALSE,FALSE)
	      | split [x1] = (TRUE, ithvar x1) (* U vs, L vs *)
	      | split x  = let val yl = length x div 2
			       val y = List.take (x,yl)
			       val z = List.drop (x,yl)
			       val (Uy,Ly) = split y
			       val (Uz,Lz) = split z
			       val Ux = AND (AND(Uy,Uz), NOT(AND(Ly,Lz)))
			       val Lx = OR (Ly,Lz)
			   in (Ux,Lx) end
	in AND (split vs) end


  (* strong support stuff *)

  fun is_strong_ (bdd :bdd) (i :int) :bool * bool (* res, assgn *) = 
	if equal FALSE (bdd.restrict bdd (bdd.assignment [(i,true )])) 
		then (true,false)
	else if equal FALSE (bdd.restrict bdd (bdd.assignment [(i,false)])) 
		then (true,true)
	else (false,false)

  fun is_strong bdd v = #1 (is_strong_ bdd v)

  (* restrict bdd only to not-strong part of support 
     returns the eliminated part of (strong) support *)

  fun weaken_support (bdd :bdd) :bdd * assignment =
	let fun f (i,(bdd,is)) = 
		let val (strong,assgn) = is_strong_ bdd i 
		in 
			if strong then 
				(exist (makeset [i]) bdd, 
				 (i,assgn)::is)
			else (bdd,is)
		end
	    val (bdd,assgn) = Vector.foldl f (bdd,[]) 
				(scanset(support bdd))
	in (bdd, assignment assgn) end
		

end


